/*
Copyright 2012 Jun Wako <wakojun@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 * scan matrix
 */
#include <stdint.h>
#include <stdbool.h>
#include <avr/io.h>
#include <util/delay.h>
#include "print.h"
#include "debug.h"
#include "util.h"
#include "matrix.h"


#ifndef DEBOUNCE
#   define DEBOUNCE	5
#endif
static uint8_t debouncing = DEBOUNCE;

/* matrix state(1:on, 0:off) */
static matrix_row_t matrix[MATRIX_ROWS];
static matrix_row_t matrix_debouncing[MATRIX_ROWS];

static matrix_row_t read_cols(void);
static void init_cols(void);
static void unselect_rows(void);
static void select_row(uint8_t row);


void matrix_init(void)
{
    debug_enable = true;
    debug_matrix = true;
    debug_mouse = true;
    // initialize row and col
    unselect_rows();
    init_cols();

    // initialize matrix state: all keys off
    for (uint8_t i=0; i < MATRIX_ROWS; i++) {
        matrix[i] = 0;
        matrix_debouncing[i] = 0;
    }
}

uint8_t matrix_scan(void)
{
    for (uint8_t i = 0; i < MATRIX_ROWS; i++) {
        select_row(i);
        _delay_us(30);  // without this wait read unstable value.
        matrix_row_t cols = read_cols();
        if (matrix_debouncing[i] != cols) {
            matrix_debouncing[i] = cols;
            if (debouncing) {
                debug("bounce!: "); debug_hex(debouncing); debug("\n");
            }
            debouncing = DEBOUNCE;
        }
        unselect_rows();
    }

    if (debouncing) {
        if (--debouncing) {
            _delay_ms(1);
        } else {
            for (uint8_t i = 0; i < MATRIX_ROWS; i++) {
                matrix[i] = matrix_debouncing[i];
            }
        }
    }

    return 1;
}

inline
matrix_row_t matrix_get_row(uint8_t row)
{
    return matrix[row];
}

/* Column pin configuration
 * col: 0
 * pin: B0
 */
static void  init_cols(void)
{
    // Input with pull-up
    DDRD  &= ~(0b11111100); //D2..D7
    PORTD |=  (0b11111100);
	DDRB  &= ~(0b00111111); //B0..B5
    PORTB |=  (0b00111111);
	DDRF  &= ~(0b00000111); //F0..F2
    PORTF |=  (0b00000111);
	DDRC  &= ~(0b00001111); //C0..C3
    PORTC |=  (0b00001111);
}

static matrix_row_t read_cols(void)
{
    return (PIND&(1<<2) ? 0 : (1<<0)) | //0: D2
           (PIND&(1<<3) ? 0 : (1<<1)) | //1: D3
           (PINB&(1<<5) ? 0 : (1<<2)) | //2: B5
           (PINB&(1<<4) ? 0 : (1<<3)) | //3: B4
           (PIND&(1<<4) ? 0 : (1<<4)) | //4: D4
           (PIND&(1<<5) ? 0 : (1<<5)) | //5: D5
           (PINB&(1<<3) ? 0 : (1<<6)) | //6: B3
           (PINB&(1<<2) ? 0 : (1<<7)) | //7: B2
           (PIND&(1<<6) ? 0 : (1<<8)) | //8: D6
           (PIND&(1<<7) ? 0 : (1<<9)) | //9: D7 
           (PINB&(1<<1) ? 0 : (1<<10)) | //10: B1 
           (PINB&(1<<0) ? 0 : (1<<11)) | //11: B0
           (PINC&(1<<0) ? 0 : (1<<12)) | //12: C0
           (PINC&(1<<1) ? 0 : (1<<13)) | //13: C1 
           (PINF&(1<<0) ? 0 : (1<<14)) | //14: F0  
           (PINF&(1<<1) ? 0 : (1<<15)) | //15: F1
           (PINC&(1<<2) ? 0 : (1<<16)) | //16: C2
           (PINC&(1<<3) ? 0 : (1<<17)) | //17: C3 
           (PINF&(1<<2) ? 0 : (1<<18));  //18: F2
}

/* Row pin configuration
 * row: 0
 * pin: B1
 */
static void unselect_rows(void)
{
    // Hi-Z(DDR:0, PORT:0) to unselect
    DDRF  &= ~0b01111000;
    PORTF &= ~0b01111000;
    DDRC  &= ~0b11110000;
    PORTC &= ~0b11110000;
}

static void select_row(uint8_t row)
{
    // Output low(DDR:1, PORT:0) to select
	switch (row) 
	{
        case 0: //0: F3
            DDRF  |= (1<<3);
            PORTF &= ~(1<<3);
            break;
        case 1: //1: C4
            DDRC  |= (1<<4);
            PORTC &= ~(1<<4);
            break;
        case 2: //2: C5
            DDRC  |= (1<<5);
            PORTC &= ~(1<<5);
            break;
        case 3: //3: F4
            DDRF  |= (1<<4);
            PORTF &= ~(1<<4);
            break;
        case 4: //4: F5
            DDRF  |= (1<<5);
            PORTF &= ~(1<<5);
            break;
        case 5: //5: C6
            DDRC  |= (1<<6);
            PORTC &= ~(1<<6);
            break;
        case 6: //6: C7
            DDRC  |= (1<<7);
            PORTC &= ~(1<<7);
            break;
        case 7: //7: F6
            DDRF  |= (1<<6);
            PORTF &= ~(1<<6);
            break;
    }
}

void matrix_print(void)
{
    print("\nr/c 0123456789ABCDEF\n");
    for (uint8_t row = 0; row < MATRIX_ROWS; row++) {
        phex(row); print(": ");
        pbin_reverse16(matrix_get_row(row));
        print("\n");
    }
}